"""
Envelope Early Terminate

Library to help with early termination of runs
Here we use a strategy where we take the top k runs or top k percent of runs
and then we build up an envelope where we stop jobs where the metric doesn't get better
"""

from .util import get_numpy
from wandb.sweeps.base import EarlyTerminate


def top_k_indicies(arr, k):
    np = get_numpy()
    return np.argpartition(arr, -k)[-k:]


def cumulative_min(history, envelope_len):
    np = get_numpy()
    cur_min = np.inf
    cum_min = []
    for j in range(envelope_len):
        if j < len(history):
            val = history[j]
        else:
            val = np.nan
        cur_min = min(cur_min, val)
        cum_min.append(cur_min)
    return cum_min


def histories_for_top_n(histories, metrics, n=3):
    np = get_numpy()
    metrics = np.array(metrics)
    histories = np.array(histories)
    indices = top_k_indicies(-metrics, n)
    top_n_histories = []
    for index in indices:
        top_n_histories.append(histories[index])
    return top_n_histories


def envelope_from_histories(histories, envelope_len):
    np = get_numpy()
    envelope = []
    cum_min_hs = []
    longest = 0
    for h in histories:
        cum_min_hs.append(cumulative_min(h, envelope_len))
    for jj in range(envelope_len):
        prev_max = -np.inf
        if (len(envelope) > 0):
            prev_max = max(envelope)
        envelope.append(max([h[jj] for h in cum_min_hs]))
    return envelope


def envelope_from_top_n(histories, metrics, n):
    histories = histories_for_top_n(histories, metrics, n)
    envelope_len = max([len(h) for h in histories])
    return envelope_from_histories(histories, envelope_len)


def is_inside_envelope(history, envelope, ignore_first_n_iters=0):
    if len(history) <= ignore_first_n_iters:
        return True

    min_val = min(history)
    cur_iter = len(history) - 1
    if cur_iter >= len(envelope):
        cur_iter = len(envelope) - 1
    return min_val < envelope[cur_iter]


class EnvelopeEarlyTerminate(EarlyTerminate):
    def __init__(self, fraction=0.3, min_runs=3, start_iter=3):
        self.fraction = fraction
        self.min_runs = min_runs
        self.start_iter = start_iter

    @classmethod
    def init_from_config(cls, config):
        pass

    def stop_runs(self, sweep_config, runs):
        np = get_numpy()
        info = {}
        terminate_run_names = []
        self._load_metric_name_and_goal(sweep_config)

        complete_run_histories = []
        complete_run_metrics = []
        for run in runs:
            if run.state == "finished":  # complete run
                history = self._load_run_metric_history(run)
                if len(history) > 0:
                    complete_run_histories.append(history)
                    complete_run_metrics.append(min(history))

        complete_runs_count = len(complete_run_histories)
        if complete_runs_count < self.min_runs:
            return [], info

        n = max(int(np.ceil(complete_runs_count * self.fraction)), self.min_runs)

        envelope = envelope_from_top_n(
            complete_run_histories, complete_run_metrics, n)

        for run in runs:
            if run.state == "running":
                history = self._load_run_metric_history(run)

                if not is_inside_envelope(history, envelope,
                                          ignore_first_n_iters=self.start_iter):
                    terminate_run_names.append(run.name)
        return terminate_run_names, info



# -*- coding: utf-8 -*-
"""
Wandb is a library to help track machine learning experiments.

For more information on wandb see https://docs.wandb.com.

The most commonly used functions/objects are:
- wandb.init — initialize a new run at the top of your training script
- wandb.config — track hyperparameters
- wandb.log — log metrics over time within your training loop
- wandb.save — save files in association with your run, like model weights
- wandb.restore — restore the state of your code when you ran a given run

For examples usage, see github.com/wandb/examples
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__version__ = "0.10.32"

# Used with pypi checks and other messages related to pip
_wandb_module = "wandb"

import sys

from wandb.errors import Error

# This needs to be early as other modules call it.
from wandb.errors.term import termsetup, termlog, termerror, termwarn

PY3 = sys.version_info.major == 3 and sys.version_info.minor >= 6
TYPE_CHECKING = False  # type: bool
if PY3:
    TYPE_CHECKING = True
    from wandb import sdk as wandb_sdk
else:
    from wandb import sdk_py27 as wandb_sdk

import wandb

wandb.wandb_lib = wandb_sdk.lib

init = wandb_sdk.init
setup = wandb_sdk.setup
save = wandb_sdk.save
watch = wandb_sdk.watch
unwatch = wandb_sdk.unwatch
finish = wandb_sdk.finish
join = finish
login = wandb_sdk.login
helper = wandb_sdk.helper
Artifact = wandb_sdk.Artifact
AlertLevel = wandb_sdk.AlertLevel
Settings = wandb_sdk.Settings
Config = wandb_sdk.Config

from wandb.apis import InternalApi, PublicApi
from wandb.errors import CommError, UsageError

_preinit = wandb_lib.preinit
_lazyloader = wandb_lib.lazyloader
from wandb import wandb_torch
from wandb import util

# Move this (keras.__init__ expects it at top level)
from wandb.data_types import Graph
from wandb.data_types import Image
from wandb.data_types import Plotly

# from wandb.data_types import Bokeh # keeping out of top level for now since Bokeh plots have poor UI
from wandb.data_types import Video
from wandb.data_types import Audio
from wandb.data_types import Table
from wandb.data_types import Html
from wandb.data_types import Object3D
from wandb.data_types import Molecule
from wandb.data_types import Histogram
from wandb.data_types import Classes
from wandb.data_types import JoinedTable

from wandb.wandb_agent import agent
from wandb.wandb_controller import sweep, controller

from wandb import superagent

# from wandb.core import *
from wandb.viz import visualize
from wandb import plot
from wandb import plots  # deprecating this
from wandb.integration.sagemaker import sagemaker_auth


# Used to make sure we don't use some code in the incorrect process context
_IS_INTERNAL_PROCESS = False


def _set_internal_process(disable=False):
    global _IS_INTERNAL_PROCESS
    if _IS_INTERNAL_PROCESS is None:
        return
    if disable:
        _IS_INTERNAL_PROCESS = None
        return
    _IS_INTERNAL_PROCESS = True


def _assert_is_internal_process():
    if _IS_INTERNAL_PROCESS is None:
        return
    assert _IS_INTERNAL_PROCESS


def _assert_is_user_process():
    if _IS_INTERNAL_PROCESS is None:
        return
    assert not _IS_INTERNAL_PROCESS


# toplevel:
# save()
# restore()
# login()
# sweep()
# agent()

# globals
Api = PublicApi
api = InternalApi()
run = None
config = _preinit.PreInitCallable(
    _preinit.PreInitObject("wandb.config"), wandb_sdk.wandb_config.Config
)
summary = _preinit.PreInitCallable(
    _preinit.PreInitObject("wandb.summary"), wandb_sdk.wandb_summary.Summary
)
log = _preinit.PreInitCallable("wandb.log", wandb_sdk.wandb_run.Run.log)
save = _preinit.PreInitCallable("wandb.save", wandb_sdk.wandb_run.Run.save)
restore = wandb_sdk.wandb_run.restore
use_artifact = _preinit.PreInitCallable(
    "wandb.use_artifact", wandb_sdk.wandb_run.Run.use_artifact
)
log_artifact = _preinit.PreInitCallable(
    "wandb.log_artifact", wandb_sdk.wandb_run.Run.log_artifact
)
define_metric = _preinit.PreInitCallable(
    "wandb.define_metric", wandb_sdk.wandb_run.Run.define_metric
)

mark_preempting = _preinit.PreInitCallable(
    "wandb.mark_preempting", wandb_sdk.wandb_run.Run.mark_preempting
)

plot_table = _preinit.PreInitCallable(
    "wandb.plot_table", wandb_sdk.wandb_run.Run.plot_table
)
alert = _preinit.PreInitCallable("wandb.alert", wandb_sdk.wandb_run.Run.alert)

# record of patched libraries
patched = {"tensorboard": [], "keras": [], "gym": []}

keras = _lazyloader.LazyLoader("wandb.keras", globals(), "wandb.integration.keras")
sklearn = _lazyloader.LazyLoader("wandb.sklearn", globals(), "wandb.sklearn")
tensorflow = _lazyloader.LazyLoader(
    "wandb.tensorflow", globals(), "wandb.integration.tensorflow"
)
xgboost = _lazyloader.LazyLoader(
    "wandb.xgboost", globals(), "wandb.integration.xgboost"
)
tensorboard = _lazyloader.LazyLoader(
    "wandb.tensorboard", globals(), "wandb.integration.tensorboard"
)
gym = _lazyloader.LazyLoader("wandb.gym", globals(), "wandb.integration.gym")
lightgbm = _lazyloader.LazyLoader(
    "wandb.lightgbm", globals(), "wandb.integration.lightgbm"
)
docker = _lazyloader.LazyLoader("wandb.docker", globals(), "wandb.docker")
jupyter = _lazyloader.LazyLoader("wandb.jupyter", globals(), "wandb.jupyter")
sacred = _lazyloader.LazyLoader("wandb.sacred", globals(), "wandb.integration.sacred")


def ensure_configured():
    global api
    api = InternalApi()


def set_trace():
    import pdb  # TODO: support other debuggers

    #  frame = sys._getframe().f_back
    pdb.set_trace()  # TODO: pass the parent stack...


__all__ = [
    "__version__",
    "init",
    "setup",
    "save",
    "sweep",
    "agent",
    "config",
    "log",
    "summary",
    "join",
    "Api",
    "Graph",
    "Image",
    "Plotly",
    "Video",
    "Audio",
    "Table",
    "Html",
    "Object3D",
    "Molecule",
    "Histogram",
]
